/*
  -------------------- file back2int.c
  Created 21 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Interface between application, and internal "back2" search routines.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* #define DEBUG_PI */
/* #define DEBUG */
#include "our_common.h"
#include "back2int.h"
#include "region.h"
#include "../libinclude/ntup.h"
#include "../libinclude/tuputil.h"
#include "../general/coord.h"
#include "../general/run.h"
#include "../general/pi.h"
#include "../general/teval.h"
#include "../general/tmask.h"
#include "generate.h"
#include "prog/memalloc.h"
#include "prog/running.h"
#include "prog/vector.h"

static FILE *s_f;
static int s_update_interval;
static int s_back_action;
static int s_back_print;
static int s_n;
static int s_tuple_kind;
static double s_max_nodes;
static double s_timeslice_nodes;
static int s_max_width;
static unsigned s_ioflags;
static double *s_lev_nodes;   /* number of nodes at each level */

#ifdef DEBUG
static void
debug_region_list(
const char *str,
const struct coord_struct * const pi,
const struct region_struct *regions,

const int len
)
{
  int i;

  fprintf(s_f,"------ %s\n", str);
  for (i = 0; i < len; i++) {
    const struct region_struct *r = &regions[i];
    fprintf(s_f,"region %d  id=%d  nomore=%d\n", i, r->id, r->nomore);
    debug_piidc_vec(pi,"ps",r->ps,r->base_len+r->extend_len);
    debug_teval("teval", r->teval);
  }
}
#endif  /*DEBUG*/

static void
header(
const struct tuple_struct *starting_tuple,
const int p,
const int q,
const int do_maxp
)
{
  FILE * const f = s_f;
  fprintf(f,"Starting back2 search.  n=%d  p=%d  q=%d  action=",s_n,p,q);
  if (s_back_action == BACK_ACTION_COUNT)   fprintf(f, "count.\n");
  else if (do_maxp)                         fprintf(f, "maxp.\n");
  else                                      fprintf(f, "find first.\n");

  fprintf(f,"max nodes=%g  slice nodes=%g  width=%d.\n",
    s_max_nodes, s_timeslice_nodes, s_max_width);

  fprintf(f, "Printing ");
  switch (s_back_print) {
    case (BACK_PRINT_NONE):
      fprintf(f, "no solution.\n");
      break;
    case (BACK_PRINT_FIRST):
      fprintf(f, "first solution.\n");
      break;
    case (BACK_PRINT_ALL):
      fprintf(f, "canonical solutions (note: 'print all' was specified).\n");
      break;
    case (BACK_PRINT_CANON_ONLY):
      fprintf(f, "canonical solutions.\n");
      break;
  }
  if (starting_tuple)
    fprintf(f, " Starting tuple specified; ignoring it.\n");
}

/*  Sets *status_ret to WORKFUNC_XX
    If we found a solution, we put the solution in *sol_ret.

    Assumes that the caller has set up *sol_ret to hold a tuple
    of size n.
*/
static void
back2_work(
const int original_p,
const int q,
const int do_maxp,              /* should we find max p for given q? */
const struct tuple_struct * const starting_tuple,   /* NULL indicates none */
                                                    /* we ignore */
const struct coord_struct * const pi,
const struct tmask_struct * const pos_tmask,
const struct tmask_struct * const neg_tmask,
struct tuple_struct * const sol_ret,
int * const status_ret
)
{
  int p;                          /* current p we are working on */
  struct region_struct *region_list;
  int nregions;                   /* current length of region list */
  int executing;                  /* region currently scheduled */
  int found_any;                  /* did we find any solutions yet? */
  double stats_tot_nodes;         /* total search nodes so far */
  double p_nodes;                 /* total nodes for given p */
  double stats_psolutions;        /* number of solutions for given p so far */
  const int verbose = (s_back_action == BACK_ACTION_COUNT) ||
                      (s_update_interval != UPDATE_NONE);
  const int loop_verbose =  (s_back_action == BACK_ACTION_COUNT) ||
                            (s_update_interval == UPDATE_OFTEN);
  const int strategic_verbose = (s_update_interval == UPDATE_STRATEGIC);
  struct to_region_struct to_region[1];
  struct from_region_struct from_region[1];
  int newp;                       /* did value of p just change? */
  int nactive_regions;        /* search regions 0 .. nactive_regions-1 */

  bsbegin();

  ASSERT(s_n>1);

  stats_tot_nodes = 0;          /* no nodes examined so far */
  stats_psolutions = 0;         /* no solutions found for given value of p */

  if (verbose || strategic_verbose)
    header(starting_tuple,original_p,q,do_maxp);

  /*  Initialize the update timer, if necessary.  Time-based
      updates, if any, emanate from search_region().
  */
  if (s_update_interval != UPDATE_NONE)
    update_init();
  if (s_update_interval > 0) {
    update_set_interval(s_update_interval);
    update_start_timer();
  }

#if defined(DEBUG_PI) || defined(DEBUG)
  fprintf(s_f,"pi:\n");
  debug_coord(pi);
#endif

  s_lev_nodes = vec_d_alloc(s_n, bsalloc);

  /*  Generate the partial solutions of length 2.
  */
  teval_setup_invariants(s_n,pos_tmask,neg_tmask,bsalloc);
  back2_gen_initial_regions(pi,s_tuple_kind,
    s_n,&region_list,&nregions,bsalloc);
  if (verbose)
    fprintf(s_f, "%d regions\n", nregions);
  p = original_p;         /* used after noneexist */
  if (nregions == 0)
    goto noneexist;

  /*  Main loop.  Picks a region, then calls the function that
      executes the search for a given region.
  */
  p_nodes                       = 0;
  to_region->pi                 = pi;
  to_region->p                  = p;
  to_region->q                  = q;
  to_region->n                  = s_n;
  to_region->tuple_kind         = s_tuple_kind;
  to_region->width              = s_max_width;
  to_region->found_first        = 0;
  to_region->update_interval    = s_update_interval;
  to_region->f                  = s_f;
  to_region->back_print         = s_back_print;
  to_region->ioflags            = s_ioflags;
  to_region->lev_nodes          = s_lev_nodes;
  found_any                     = 0;          /* no solutions found yet */
  newp                          = 1;
  DO_GCCHAPPY(executing = nactive_regions = -1;)

  while (1) {
    if (newp) {
      int i;

      if (strategic_verbose)
        fprintf(s_f, "Beginning work on p=%d, q=%d.\n", p,q);

      vec_d_zero(s_lev_nodes, s_n);
      teval_setup_pq(p,q);              /* must be done before region sort */
      back2_region_sort(region_list,&nregions);
      if (nregions == 0) {
        if (found_any) goto success;
        else           goto noneexist;
      }
      if (s_max_width == 0) nactive_regions = nregions;
      else                  nactive_regions = MIN(s_max_width,nregions);
      newp = 0;
      executing = 0;
      /*  The order of the candidate lists may change when p changes,
          because the evaluation functions change.  At a given level,
          we want to guarantee that we have carried out the search
          for candidates with lower index than the index of the current
          partial solution.  This cannot be guaranteed, however, since
          the order of the candidates may change when p changes.  So
          we restart the search for all regions.

          This is also a good time to generate a (possibly new) random
          seed for each search region.
      */
      for (i = 0; i < nregions; i++) {
        long random();
        const long r = random();

        region_list[i].extend_len = 0;
        ASSERT(r >= 0);               /* region seed is unsigned */
        region_list[i].seed = r;
      }
    }

    if (loop_verbose) {
      fprintf(s_f, "\n");
      fprintf(s_f,
              "p=%d  q=%d  %d regions, %d active.  Working on region %d:\n",
              p,q,nregions,nactive_regions,region_list[executing].id);
      ps_print(s_f,pi,
                  region_list[executing].ps,region_list[executing].base_len,1);
      fprintf(s_f, "%g psolutions, %g nodes, %g nodes for current p.\n",
              stats_psolutions, stats_tot_nodes, p_nodes);
    }
    D(debug_region_list("top of control loop",pi,region_list,nactive_regions);)

    to_region->found_first     = (stats_psolutions > 0);
    to_region->p_nodes         = p_nodes;
    to_region->nactive_regions = nactive_regions;
    {
      const double slice  = s_timeslice_nodes;
      const double maxn   = s_max_nodes;
      if (maxn == 0)      /* handles maxn==0 && slice==0 */
        to_region->max_slice_nodes = slice;
      else if (slice == 0) {
        ASSERT(maxn > p_nodes);
        to_region->max_slice_nodes = maxn - p_nodes;
      }
      else if (p_nodes + slice > maxn) {
        ASSERT(maxn > p_nodes);
        to_region->max_slice_nodes = maxn - p_nodes;
      }
      else
        to_region->max_slice_nodes = slice;
    }
    back2_search_region(to_region, &region_list[executing],
                        from_region, sol_ret->pixels);
    p_nodes           += from_region->nodes;
    stats_tot_nodes   += from_region->nodes;
    stats_psolutions  += from_region->solutions;
    if (loop_verbose) {
      fprintf(s_f, "Returned from working on region %d.\n",
        region_list[executing].id);
      fprintf(s_f, "Examined %g nodes, found %g solutions.\n",
        from_region->nodes, from_region->solutions);
    }

    switch (from_region->status) {
    case SEARCH_INCONCLUSIVE:
      executing = (executing+1) % nactive_regions;
      break;
    case SEARCH_EXHAUSTED:
      if (loop_verbose)
        fprintf(s_f, "Region exhausted.\n");
      if (nactive_regions == 1) {   /* region exhausted was last region */
        if (found_any) goto success;
        else           goto noneexist;
      }
      if (executing < nregions-1) {
        memmove(  (char *) & region_list[executing],
                  (char *) & region_list[executing+1],
                  sizeof(region_list[0]) * (nregions-executing-1) );
          /* note: allocated fields of old region_list[executing] are lost.  */
      }
      ASSERT(executing < nactive_regions);
      if (executing == nactive_regions-1)
        executing = 0;
      nactive_regions--;
      nregions--;
      break;
    case SEARCH_FOUND_FIRST:
      ASSERT(sol_ret->n == s_n);
      found_any = 1;
      sol_ret->p = p;
      if (s_back_action == BACK_ACTION_COUNT) {}
      else if (do_maxp) {
        ASSERT(to_region->found_first == 0);
        if (s_ioflags & IOF_REPORT_RESULTS) {
          lev_nodes_print(s_f,s_lev_nodes,s_n,p,q);
          vec_d_zero(s_lev_nodes, s_n);
        }
        for (p++; tmask_pos_satisfied(sol_ret->pixels,sol_ret->n,p,pos_tmask);
                                                                p++)  {}
        sol_ret->p = p-1;
        to_region->p = p;
        stats_psolutions = 0;
        p_nodes = 0;
        newp = 1;

        if (loop_verbose)
          fprintf(s_f, "Solution found.  Set p=%d.\n", p);
      }
      else
        goto success;
      break;
    default:
      fatal("back2_work()");
    }   /* switch */

    if ((s_max_nodes > 0) && (p_nodes >= s_max_nodes)) {
      if (verbose)
        fprintf(s_f, "Maximum number of nodes reached; exiting.\n");

      if (found_any)
        goto success;
      goto notfound;
    }
  }         /* main loop */

notfound:
  *status_ret = WORKFUNC_NOTFOUND;
  goto done;

noneexist:
  *status_ret = WORKFUNC_NONEEXIST;
  goto done;

success:
  sol_ret->q = q;
  if (s_back_action == BACK_ACTION_COUNT)   *status_ret = WORKFUNC_P;
  else if (do_maxp)                         *status_ret = WORKFUNC_PMAX;
  else                                      *status_ret = WORKFUNC_P;
  goto done;


done:

  if (verbose || strategic_verbose) {
    fprintf(s_f, "\n");
    fprintf(s_f, "Back2 search completed.  %g nodes.", stats_tot_nodes);
    if (s_back_action == BACK_ACTION_COUNT)
      fprintf(s_f, "  %g solutions.", stats_psolutions);
    fprintf(s_f, "\n");
  }

  if (s_ioflags & IOF_REPORT_RESULTS)
    lev_nodes_print(s_f,s_lev_nodes,s_n,p,q);

  bsend();
}

/*
  Interface between user and back2 search functions.
  This is what the user calls.
  It is declared in ../libinclude/ntup.h, and not back2int.h as it would be
  if we following the usual convention of placing the function declaration
  in the .h file corresponding to the .c file in which the function resides.

  This function should free all memory it allocates if it sets
  the status field to anything other than NTUP_OK.
*/
void
ntup_find_back2(
const struct to_ntup_struct *to,
struct from_ntup_struct *from,
const int     back_action,          /* BACK_ACTION_XX */
const int     back_print,           /* BACK_PRINT_XX */
/* maximum number of nodes before stopping */
const double  back_max_nodes,       /* must be 0 or more; 0 means infinity */
const double  timeslice_nodes,      /* must be 0 or more; 0 means infinity */
const int     max_width             /* restricted backtracking, 0=infinity */
)
{
  D(fprintf(s_f,"---------------- entering ntup_find_back2\n");)

  /*  Initialize bit manipulation package.  bit_global_init() does
      virtually nothing if it has been called previously.
  */
  bit_global_init();

  /*  -------------------------- Check input for errors, report
                                 input parameters if reporting is specified.
   */
  ntup_check_input(to, & from->status);
  if (from->status == NTUP_OK) {
    if (timeslice_nodes < 0)
      from->status = NTUP_INVALID_BACK2_TIMESLICE_NODES;
    if (to->n < 3)
      from->status = NTUP_INVALID_BACK2_N;
    if (max_width < 0)
      from->status = NTUP_INVALID_BACK2_RESTRICT;
    ntup_check_input_back(back_action,back_print,back_max_nodes,& from->status);
  }

  if (from->status != NTUP_OK)  return;

  /* Store data needed for future use.
  */
  s_f                 = to->f;
  s_ioflags           = to->ioflags;
  s_update_interval   = to->update_interval;
  s_back_action       = back_action;
  s_back_print        = back_print;
  s_max_nodes         = back_max_nodes;
  s_n                 = to->n;
  s_tuple_kind        = to->tuple_kind;
  s_timeslice_nodes   = timeslice_nodes;
  s_max_width         = max_width;

  if (s_ioflags & IOF_REPORT_INPUT) {
    fprintf(s_f,"---------- Begin report of back2 input.\n");
    if (back_max_nodes == 0)
      fprintf(s_f, "max_nodes: infinity   ");
    else
      fprintf(s_f, "max_nodes: %g   ", back_max_nodes);
    if (timeslice_nodes == 0)
      fprintf(s_f, "slice_nodes: infinity   ");
    else
      fprintf(s_f, "slice nodes: %g   ", timeslice_nodes);
    if (max_width == 0)
      fprintf(s_f, "width: infinity\n");
    else
      fprintf(s_f, "width: %d\n", max_width);
    report_back(s_f,back_action,back_print);
    report_input(to);
    fprintf(s_f,"---------- End report of back2 input.\n\n");
  }

  /* --------------------------------
      Call back2_work() with p and q as needed, keep track of solutions.
  */
  ntup_run(to,1,from,back2_work);

  D(fprintf(s_f,"---------------- exiting ntup_find_back2\n");)
}
