/*
  -------------------- file argproc.c
  Created 6 Oct 92 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Routines for processing command-line arguments.

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "our_common.h"
#include "argproc.h"
#include "io.h"
#include "memalloc.h"
#include <limits.h>

extern void usage P(( const char * ));


void
badarg(
const char *label,  /* print this string before the argument */
const char *arg     /* argv[an], where an is the argument with the problem */
)
{
  char *buff;
  buff = xmalloc(strlen(arg)+strlen(label)+128);
  sprintf(buff, "%s  Detected at argument '%s'.\n", label, arg);
  usage(buff);
  xfree(buff);  /* in case usage() doesn't exit() (it's supposed to, though) */
}

/* Use this routine to see if an argument is "argname x(1) x(2) .. x(n)",
  where each x(i) is a value that should fit into an "int".
  If the argument is argname, then we get the values of x.  Each
  value must be in a specified range.  (Use min=INT_MIN and max=INT_MAX
  if you want no range testing.)  The caller can do additional range
  testing if he wants.

  Return  1 if argument was successful "argname ..."
          0 if argument was not "argname" or there are no more arguments
          fatal usage message if argument was unsuccessful "argname ..."
  
  Assumes the existence of the usual function "usage( const char * )"
  and the usual variable "char *Progname".

  Prevents you from using the same argument twice (you should initialize
  *already to 0 before your argument processing loop).
*/
int
argint_vec(
const char *argname,
const int n,
const int min,
const int max,
const int argc,
char      **argv,
/* return variables */
int       *arg_pos,
int       *num_ret,   /* vector is returned in num_ret[0] .. num_ret[n-1] */
int       *already    /* specified this argument yet? */
)
{
  const int a1 = *arg_pos;
  int i;
  extern char *Progname;

  if (a1 >= argc)  return(0);
  if (strcmp(argv[a1], argname))  return(0);
  if (a1+n >= argc) {
    fprintf(stderr, "%s: Missing part of '%s' arg.\n", Progname, argname);
    goto fail;
  }
  if (already && *already) {
    fprintf(stderr, "%s: Extra specification of '%s' arg.\n",
      Progname, argname);
    goto fail;
  }
  for (i = 0; i < n; i++) {
    int status;
    our_atoi(argv[a1+1+i], &status, & num_ret[i]);
    if (status) {
      fprintf(stderr, "%s: component %d of '%s' arg must be an integer.\n",
        Progname, i+1, argname);
      goto fail;
    }
    if ((num_ret[i] < min) || (num_ret[i] > max)) {
      fprintf(stderr,
        "%s: component %d of '%s' arg must be in range %d .. %d\n",
        Progname, i+1, argname, min, max);
      goto fail;
    }
  }

  (*arg_pos) += n+1;
  if (already)
    (*already) = 1;
  return(1);

  fail:
    usage("Error on command line.");
#ifdef  GCCHAPPY
  return(-1);
#endif  /*GCCHAPPY*/
}


/*  Interface to argint_vec when the vector is length 1
*/
int
argint(
const char *argname,
const int min,
const int max,
const int argc,
char      **argv,
/* return variables */
int       *arg_pos,
int       *num_ret,
int       *already    /* specified this argument yet? */
)
{
  return(argint_vec(argname,1,min,max,argc,argv,arg_pos,num_ret,already));
}

/* Use this routine to see if an argument is "argname x", where x is
  a a string, and if it is, get the value of x.

  Return 1 if argument was successful "argname x"
          0 if argument was not "argname" or there are no more arguments
  
  Assumes the existence of the usual function "usage( const char * )"
  and the usual variable "char *Progname".

  Prevents you from using the same argument twice (you should initialize
  *already to 0 before your argument processing loop).
*/

int
argstr(
const char *argname,
const int argc,
char      **argv,
/* return variables */
int       *arg_pos,
char      **str,      /* if non-NULL, we make *str point to string argument */
int       *already    /* specified this argument yet? */
)
{
  const int a1 = *arg_pos;
  const int a2 = a1+1;
  extern char *Progname;

  if (a1 >= argc) return(0);
  if (strcmp(argv[a1], argname))
    return(0);
  if (a2 >= argc) {
    fprintf(stderr, "%s: Missing second part of '%s' arg.\n",
      Progname, argname);
    goto fail;
  }
  if (already && *already) {
    fprintf(stderr, "%s: Specified ('%s' or related arg) twice.\n",
			Progname, argname);
    goto fail;
  }
  *str = argv[a2];

  (*arg_pos) += 2;
  if (already)
    (*already) = 1;
  return(1);

  fail:
    usage("Error on command line.");
#ifdef  GCCHAPPY
  return(-1);
#endif  /*GCCHAPPY*/
}


/*  Just like argint, except that there is no second arg.  This is
    good for simple arg like "-O2".
*/
int
argnull(
const char *argname,
const int argc,
char      **argv,
/* return variables */
int       *arg_pos,
int       *already    /* specified this argument yet? can be NULL*/
)
{
  const int a1 = *arg_pos;
  extern char *Progname;

  if (a1 >= argc)  return(0);
  if (strcmp(argv[a1], argname))
    return(0);
  if (already && *already) {
    fprintf(stderr, "%s: Specified ('%s' or related arg) twice.\n",
			Progname, argname);
    usage("Error on command line.");
  }
  (*arg_pos)++;
  if (already)
    (*already) = 1;
  return(1);
}

/*
  Call this function when it is known that the next argument(s)
  specify a computation method.

  Currently accepts:
    -linear p q r           yields ARG_COMPUTE_LINEAR
  TODO: make it accept a polynomial with rational coefficients
  Example: you have an argument -reg compute, where -reg means to
  generate a regular graph, and compute says how to compute the
  degree of the graph from n.
  
  Generates a fatal usage message if the denominator is zero, or if the
  ratio is outside the specified range.
  
  If there was an error, prints error message and calls usage()

  Return  1 if argument was successful "argname ..."
          0 if argument was not "argname"
          fatal usage message if argument was unsuccessful "argname ..."
*/
int
argcompute(
const char *argname,
const int argc,
char **argv,
const double *lowbound, /* if non-NULL, require rat>=lowbound[0]/lowbound[1] */
const double *upbound,  /* if non-NULL, require rat<=upbound[0]/upbound[1] */
/* return variables */
struct arg_compute_struct *c,
int *arg_pos,
int *already            /* specified this argument yet? */
)
{
  /*TODO - process command line args as though they were doubles */
  int buff[3];
  int dummy;

  dummy = 0;
  if (! argnull(argname,argc,argv,arg_pos,&dummy))  return(0);
  dummy = 0;
  if (argint_vec("-linear",3,INT_MIN,INT_MAX,argc,argv,arg_pos,buff,&dummy)) {
    const int p = buff[0];
    const int q = buff[1];
    if (q == 0)
      {  fprintf(stderr, "Denominator in ratio cannot be 0.\n");  goto fail;  }
    if (lowbound) {
      if (p*lowbound[1] < lowbound[0]*q)
        {  fprintf(stderr, "Ratio is too small.\n");  goto fail;  }
    }
    if (upbound) {
      if (p*upbound[1]> upbound[0]*q)
        {  fprintf(stderr, "Ratio is too large.\n");  goto fail;  }
    }
    c->kind = ARG_COMPUTE_LINEAR;
    c->r[0] = p;
    c->r[1] = q;
    c->s = buff[2];
  }
  else  return(0);

  if (already)
    (*already) = 1;
  return(1);

fail:
  usage("Error on command line.");
  DO_GCCHAPPY(return(1);)
}

double
compute_arg_do(
const struct arg_compute_struct *c,
const double base
)
{
  switch (c->kind) {
  case ARG_COMPUTE_LINEAR:
    return( ((c->r[0] * base) / c->r[1]) + c->s);
  default:
    fatal("compute_argvalue");
    DO_GCCHAPPY(return(0.0);)
  }
}

void
compute_arg_print(
FILE *f,
const struct arg_compute_struct *c,
const char *str     /* the base value */
)
{
  switch (c->kind) {
  case ARG_COMPUTE_LINEAR: {
    const double p = c->r[0];
    const double q = c->r[1];
    const double s = c->s;
    if (p == 0)
      fprintf(f,"%g", s);
    else {
      /* ratio */
      if (p == q)  fprintf(f,"%s",str);
      else {
        if (q == 1) fprintf(f,"%s * %g", str, p);
        else        fprintf(f,"%s * (%g/%g)", str, p, q);
      }
      /* sum */
      if (s < 0)      fprintf(f," - %g", -1 * s);
      else if (s > 0) fprintf(f," + %g", s);
    }
    break;
  }
  default:
    fatal("compute_argvalue");
  }
}

void
compute_arg_set_linear(
struct arg_compute_struct *c,
const double p,
const double q,
const double s
)
{
  c->kind = ARG_COMPUTE_LINEAR;
  c->r[0] = p;
  c->r[1] = q;
  c->s = s;
}
