/*
  -------------------- file bitpar.h
  Created 17 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Functions regarding "bit-parallel" representation.  An l-level bit
  parallel representation is an array.  It has l rows.  Each row of the
  array is a word.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



/*
  See the comments before unsigned_to_bitpar() for more.

  These routines depend on the fact that the kth most significant bit
  of an unsigned int value "v" can be obtained with (1u << (k-1)).
  I think according to the ANSI C standard that this is not
  MACHINE DEPENDENT, but I have not checked.  TODO.

  USERS MUST CALL bitword_global_init() BEFORE CALLING THESE FUNCTIONS.
*/

#ifndef _180594_143448
#define _180594_143448

#include "bitword.h"

int bitpar_levels_needed P(( const BITWORD ));

/**************************************************************************
  inline functions (be careful about changing the order of func. definition)
**************************************************************************/

/* declarations (included for compatibility with non-gnuc compilers) */

#ifndef __GNUC__
BITWORD * bitpar_alloc P(( const int, void  *(*) P((const int)) ));
void bitpar_zero P(( BITWORD *, const int ));
void bitpar_copy P(( BITWORD *, const BITWORD *, const int ));
void unsigned_to_bitpar P(( const unsigned int, const int, BITWORD * ));
void bitpar_add_onerow P(( const BITWORD, BITWORD * ));
int bitpar_compare_max P(( const BITWORD *, const BITWORD *, const int ));
unsigned bitpar_find_max P(( const BITWORD *, const int ));
unsigned bitpar_find_max_and_maxmask
                         P(( const BITWORD *, const int, BITWORD * const  ));
BITWORD bitpar_make_mask_from_unsigned
  P(( const BITWORD * const, const int, const unsigned  ));
void bitpar_zero_columns P(( BITWORD * const, const int, const unsigned ));
int bitpar_count_columns P(( BITWORD * const, const int, const unsigned ));

#endif  /*__GNUC__*/

/*  note: inline is defined to the empty string unless we're using gnuc */
#undef  INLINE
#ifdef  SRCFILE_LIBPROG_BITPAR
#define INLINE  __inline__
#else
#define INLINE  __inline__ extern
#endif

#if defined(__GNUC__) || defined(SRCFILE_LIBPROG_BITPAR)

INLINE BITWORD *
bitpar_alloc(
const int l,                        /* number of levels */
void  *(*allocfunc) P((const int))
)
{
  return( (*allocfunc) (sizeof(BITWORD) * l) );
}

INLINE void
bitpar_zero(
BITWORD *a,
const int l
)
{
  int i;
  for (i = 0; i < l; i++)
    a[i] = ALL_BITS_ZERO;
}

INLINE void
bitpar_copy(
BITWORD *dest,
const BITWORD *src,
const int l
)
{
  int i;
  for (i = 0; i < l; i++)
    dest[i] = src[i];
}

/*  Convert a nonnegative integer v into a bit-parallel representation.
    Example:  for v=13, on a 32-bit machine, and l=4, we'd set a[] to:

      a[0]   11111111111111111111111111111111       (least significant)
      a[1]   00000000000000000000000000000000       
      a[2]   11111111111111111111111111111111     
      a[3]   11111111111111111111111111111111       (most significant)
    
    The letter "l" stands for "levels" and is used this way in other
    functions.
*/
INLINE void
unsigned_to_bitpar(
const unsigned int v, /* value to convert */
const int l,          /* number of levels to use (see also bitpar_levels) */
BITWORD *a            /* we set a[0], a[1], ..., a[l-1] */
)
{
  int i;
  const unsigned int one = 1;
  for (i = 0; i < l; i++)
    a[i] = (v & (one << i)) ? ALL_BITS_ONE : ALL_BITS_ZERO;
}

/*
    Do the parallel add of s and a.
    t             one-row parallel bits
    a[0 ..]       multi-row parallel bits

    The result is stored in a[].
    It is assumed that a[] is big enough to hold the result and that
    a is at least length 1.
*/
INLINE void
bitpar_add_onerow(
const BITWORD t,
BITWORD *a
)
{
  BITWORD *ap;      /* starts at least significant word, goes to msw */
  BITWORD c;        /* carry */

  c = a[0] & t;
  a[0] ^= t;
  if (c) {  /* do not refer to a[1] because a[1] may not exist if c is 0 */
    for (ap = & a[1]; 1; ap++) {
      const BITWORD newc = (*ap) & c;
#ifdef NODEBUG
      printf("bit_par_add_onerow: top of loop.  carry in=%x\n", c);
#endif  /*NODEBUG*/
      (*ap) ^= c;
      if (newc == 0)  break;
      c = newc;
    }
  }
}

/*  qsort-style comparison function for bitpar entities.
    We compare a and b.  We assume they each have "l" levels.
    We do the comparison as follows.  First, we interpret each column
    of a to be a nonnegative integer, and similarly for b.  So, for
    example, on a 32-bit machine, a has 32 integers, as does b.  Then we
    form A = max(a) and B = max(b), where here max is taken over the 32
    integers of a, and similarly for b.  Example (l = 3), on an 8 bit
    machine:

              a[0] = 11001100  (lsb)            b[0] = 00101010  (lsb)
              a[1] = 01010101                   b[1] = 11000011
              a[2] = 11110111  (msb)            b[2] = 11110101  (msb)
      columns of a = 57461743           columns of b = 66541436
    So A = max(5,7,4,6,1,7,4,3) = 7, B = max(6,6,5,4,1,4,3,6) = 6,
    A > B, and we return 1.

    Returns -1, 0, or 1, respectively,
    according to whether A < B, A = B, or A > B.

    Compatible with unsigned_to_bitpar().

    The caller should consider carefully whether the input operands have
    all their bits set and how this affects this computation.  For example,
    in the bit array representation (bitword.h and bitarray.h), unused bits
    are 0.

    This is a very fast implementation.
*/
INLINE int
bitpar_compare_max(
const BITWORD *a,
const BITWORD *b,
const int l       /* number of levels; l >= 0 */
)
{
  BITWORD amask, bmask;         /* 1 for columns that might be maximum */
  int i;

  for (amask = bmask = ALL_BITS_ONE, i = l-1; i >= 0; i--) {
    const BITWORD newa = amask & a[i];
    const BITWORD newb = bmask & b[i];
    if (newa) {
      if (newb)     {  amask = newa;  bmask = newb;  }
      else          return(1);
    }
    else if (newb)  return(-1);
    /* if both newa and newb are 0; leave amask and bmask alone */
  }
  return(0);
}

/*  Find the value of the maximum column, in the sense of bitpar_compare_max.
    We interpret each column of a to be a nonnegative integer.
    For example, on a 32-bit machine, a has 32 integers.  Then we
    return A = max(a).  Example (l = 3), on an 8 bit machine:

              a[0] = 10001100  (lsb)
              a[1] = 01010001
              a[2] = 11110111  (msb)
      columns of a = 56461543
    So A = max(5,6,4,6,1,5,4,3) = 6.
*/
INLINE unsigned
bitpar_find_max(
const BITWORD *a,
const int l       /* number of levels; l >= 0 */
)
{
  int i;
  int sum = 0;                    /* current partial sum for maximum */
  BITWORD amask = ALL_BITS_ONE;   /* 1 for columns that might be maximum */

  for (i = l-1; i >= 0; i--) {
    const BITWORD newa = amask & a[i];
    if (IS_BITWORD_NOT_ALL_ZERO(newa))
      {  amask = newa;  sum += (1 << i);  }
  }
  return(sum);
}

/*  Same as bitpar_find_max, but we also return a mask of columns
    that have the maximum value.
*/
INLINE unsigned
bitpar_find_max_and_maxmask(
const BITWORD *a,
const int l,       /* number of levels; l >= 0 */
BITWORD * const maxmask_ret
)
{
  int i;
  int sum = 0;                    /* current partial sum for maximum */
  BITWORD amask = ALL_BITS_ONE;   /* 1 for columns that might be maximum */

  for (i = l-1; i >= 0; i--) {
    const BITWORD newa = amask & a[i];
    if (IS_BITWORD_NOT_ALL_ZERO(newa))
      {  amask = newa;  sum += (1 << i);  }
  }
  *maxmask_ret = amask;
  return(sum);
}
  
    

/*  Make a mask that gives the columns that have the given value.
    Example:
              a[0] = 10001100   (lsb)
              a[1] = 01010001
              a[2] = 11110111   (msb)
    with v=6 would return
                     01010001
    Note: only the first "l" bits of v are used.
        
*/
INLINE BITWORD
bitpar_make_mask_from_unsigned(
const BITWORD * const a,
const int l,
const unsigned v
)
{
  BITWORD mask = ALL_BITS_ONE;
  const unsigned one = 1;
  int i;

  for (i = 0; i < l; i++) {
    if (v & (one << i))
      mask &= a[i];
    else
      mask &= (~ a[i]);
    if (IS_BITWORD_ALL_ZERO(mask))
      break;
  }
  return(mask);
}

/*  Set the bits of the columns that have value "v" to 0.
    Leave the other bits alone.  Example:
    Before:
              a[0] = 10001100   (lsb)
              a[1] = 01010001
              a[2] = 11110111   (msb)
    After the operation (with v=6):
              a[0] = 10001100
              a[1] = 00000000
              a[2] = 10100110
    Note: only the first "l" bits of v are used.
*/
INLINE void
bitpar_zero_columns(
BITWORD * const a,
const int l,
const unsigned v
)
{
  const BITWORD mask = ~ bitpar_make_mask_from_unsigned(a,l,v);
  int i;

  for (i = 0; i < l; i++)
    a[i] &= mask;
}

/*  Count the number of columns that have value "v".  Example:
              a[0] = 10001100   (lsb)
              a[1] = 01010001
              a[2] = 11110111   (msb)
    bitpar_count_columns(a,3,6) returns 3 since 3 columns have value "6".
    Note: only the first "l" bits of v are used.
*/
INLINE int
bitpar_count_columns(
const BITWORD * const a,
const int l,
const unsigned v
)
{
  BITWORD mask = bitpar_make_mask_from_unsigned(a,l,v);
  return(bitword_count(mask));
}

#endif  /* defined(__GNUC__) || defined(source file) */

#endif /* _180594_143448 */
