/*
  -------------------- file running.c
  Created 8 Jun 93 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Common tasks for running jobs - show progress reports; show banner
  with host name, date/time, PID.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "our_common.h"

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>

#include "running.h"
#include "timer.h"

static struct timeval job_tbegin;     /* time (seconds) of start */
static struct rusage job_rbegin;      /* rusage at start */

static long update_interval = -1;     /* seconds between updates */
static struct timeval update_tbegin;  /* time (seconds) of start of proc */
static struct rusage update_rbegin;   /* rusage at start */
static int did_update_init = 0;       /* did we call update_init? */
static struct timeval update_timer;   /* time (seconds) of start */

int getrusage P((int, struct rusage *));

/* Display time like Sun Sep 16 01:03:52 1973 */
void
show_datetime(
FILE *f
)
{
  time_t      clock;

  time_t      time P(( time_t * ));
  /* struct tm * localtime P(( const time_t * ));
  char *      asctime P(( const struct tm * )); */

  (void) time(&clock);
  fprintf(f, "%.24s",asctime(localtime(&clock)) );    /* no newline (only 24) */
}

/*  Display host name like "abc.cs.rpi.edu".
    Displays up to "maxlen" characters.
*/
void
show_host(
FILE *f,
const int maxlen
)
{
  int     gethostname P(( char *, int ));
  char    hostname[maxlen+1];

  if (gethostname(hostname, maxlen))
    fatal("prog/show_host: gethostname");
  hostname[maxlen] = '\0';  /* I don't know if this is necessary - AS */
  (void) fputs(hostname, f);
}

/* Print a line like this:

Sun Sep 16 01:03:52 host=01234567890123456789 pid=12345

*/
void
job_banner(
FILE *f
)
{
  pid_t   getpid P (());

  show_datetime(f);
  fprintf(f, " ");
  show_host(f,20);
  fprintf(f, " pid=%5ld", (long int) getpid());
  fprintf(f, "\n");
}

/*  Print the start banner, and initialize information for keeping track of:
        real time, user cpu, system cpu

    Print a line like this:
------------- Started Sun Sep 16 01:03:52 pid=12345 host=01234567890123456789
    with a blank line after it.

  Beware if you call this function more than once per process.
*/

void
job_start(
FILE *f
)
{
  int getpagesize P (());

  begin_timing(&job_tbegin);
  getrusage(RUSAGE_SELF, &job_rbegin);
  fprintf(f, "------------- Started ");
  job_banner(f);
  fprintf(f,"page size=%d bytes\n", getpagesize());
  fprintf(f,"\n");
}

static void
show_times(
FILE            *f,
struct timeval  *timer,
struct rusage   *rstart,
struct rusage   *rend
)
{
  long                      u_cpu_sec, u_cpu_msec;
  long                      s_cpu_sec, s_cpu_msec;
  long                      seconds,mseconds; /* sec, microseconds */

  end_timing(timer,&seconds,&mseconds);
  end_timing2(&rstart->ru_utime, &rend->ru_utime, &u_cpu_sec, &u_cpu_msec);
  end_timing2(&rstart->ru_stime, &rend->ru_stime, &s_cpu_sec, &s_cpu_msec);

  fprintf(f,"User CPU time    %10ld.%06ld s\n", u_cpu_sec, u_cpu_msec);
  fprintf(f,"System CPU time  %10ld.%06ld s\n", s_cpu_sec, s_cpu_msec);
  fprintf(f,"Real time        %10ld.%06ld s\n", seconds, mseconds);
}




/*
  Print a line like like this:
------------- Completed Sun Sep 16 01:03:52 pid=12345 host=01234567890123456789

  Displays resource usage since job_start().
*/
void
job_end(
FILE *f
)
{
  struct rusage             rend;

  int getpagesize P (());

  fprintf(f,"\n");
  fprintf(f, "------------- Completed ");
  job_banner(f);

  getrusage(RUSAGE_SELF, &rend);

  fprintf(f,"\n");
  show_times(f,&job_tbegin,&job_rbegin, &rend);
  fprintf(f,"Max. resident set       %10ld bytes (over life of process)\n",
    rend.ru_maxrss*getpagesize());
}

/* Routines for giving periodic updates.  The protocol is:

  update_init();
  update_set_interval(seconds between updates);
  update_start_timer()
  while (1) {
    ... execute some code
    if (update_test()) {
      (it's time to do an update)
      update_header_short() or update_header_long() if you want
      print your own update information if you want
      update_start_timer();             (for next update)
    }
  }
*/

void
update_init()
{
  begin_timing(&update_tbegin);
  getrusage(RUSAGE_SELF, &update_rbegin);
  did_update_init = 1;
}

/*  Change the update interval.  You can do this even if the update interval
    was set previously.
*/
void
update_set_interval(
const int interval
)
{
  if (interval < 0)   /* takes care of -1 */
    fatal("Update interval must be at least 0.");
  update_interval = interval;
}

void
update_start_timer()
{
  begin_timing(&update_timer);
}

/*  Returns 0 if it is not time to do an update.  Otherwise, returns
    the number of seconds since the update timer was started.
    
    This is not a very efficient routine.  The caller may want to take
    precautions to only call this routine every so often.  For example,
    in a loop, instead of calling this routine every time through the loop,
    it may be better to call it every 1000th time through the loop.
*/
int
update_test()
{
  long seconds, mseconds;

  if (update_interval == -1)
    fatal("Update interval not set.");
  end_timing(&update_timer,&seconds,&mseconds);
  if (seconds >= update_interval)
    return(seconds);
  else
    return(0);
}

void
update_header_short(
FILE *f
)
{
  if (! did_update_init) fatal("update header_short: missing update_init().");
  fprintf(f, "------ Update ");
  show_datetime(f);
  fprintf(f,"\n");
}

void
update_header_long(
FILE *f
)
{
  struct rusage rend;

  if (! did_update_init) fatal("update header_long: missing update_init().");
  fprintf(f, "------ Update ");
  job_banner(f);
  getrusage(RUSAGE_SELF, &rend);
  show_times(f,&update_tbegin,&update_rbegin,&rend);
}
